#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <string.h>
#include <stdint.h>
#include <SDL2/SDL.h>
#include <SDL2/SDL_video.h>
#include <cuda.h>
#include <cuda_runtime.h>
#include "universe.hpp"
#include "npp.h"
#include "custom_kernels.cuh"
#include "util.hpp"

#define DELTA_TIME 0.5


void universe::process_collisions_with_rays(void)
{
    float temp_directions[][3] = 
    {
        {1.0,0.0,0.0},
        {0.0,1.0,0.0},
        {1.0,0.0,0.0},
        {0.0,1.0,0.0}
    };

    float temp_positions[][3] = 
    {
        {0.0,0.0,0.0},
        {0.0,0.0,0.0},
        {800.0,800.0,800.0},
        {800.0,800.0,800.0}
    };

    float alpha = 1.0;
    float beta = 0.0;

    float * dev_dot_product_result;
    float * dev_direction_vec;
    
    CHECK_CUDA_ERROR(cudaMallocManaged(&(dev_dot_product_result),
                long(num_of_particles*sizeof(float))));

    CHECK_CUDA_ERROR(cudaMallocManaged(&(dev_direction_vec),
                long(num_of_particles*3*sizeof(float))));

    for(int i = 0; i < 4; i++)
    {
        //Fill scratchpad A with vec1 = line_position - particle_position
        for (int j = 0; j < 3; j++)
        {
            CHECK_CUDA_ERROR(nppsSubCRev_32f(&device_position[array_size * j],
                        temp_positions[i][j], &device_scratch_padA[num_of_particles * j],
                        num_of_particles));
        }
        alpha = 1.0;
        beta = 0.0;
        CHECK_CUDA_ERROR(cublasSgemm(cublas_handle, CUBLAS_OP_N, CUBLAS_OP_N,
                    num_of_particles, 1, 3, &alpha, device_scratch_padA,
                    num_of_particles, &device_ray_direction[i * 3], 3, &beta,
                    dev_dot_product_result, num_of_particles));

        //Use dot product to find projections onto the line directions
        CHECK_CUDA_ERROR(cublasSgemm(cublas_handle, CUBLAS_OP_N, CUBLAS_OP_T,
                    num_of_particles, 3, 1, &alpha, dev_dot_product_result,
                    num_of_particles, &device_ray_direction[i * 3], 3, &beta,
                    device_scratch_padA, num_of_particles));

        //calc norm
        CHECK_CUDA_ERROR(nppsSqr_32f(device_scratch_padA, dev_direction_vec,
                    num_of_particles * 3));

        CHECK_CUDA_ERROR(cublasSaxpy(cublas_handle, num_of_particles, &alpha,
                    &dev_direction_vec[num_of_particles * 1], 1, 
                    dev_direction_vec, 1));

        CHECK_CUDA_ERROR(cublasSaxpy(cublas_handle, num_of_particles, &alpha,
                    &dev_direction_vec[num_of_particles * 2], 1,
                    dev_direction_vec, 1));

        CHECK_CUDA_ERROR(nppsSqrt_32f_I(dev_direction_vec, num_of_particles));

        //Normalize vector
        CHECK_CUDA_ERROR(nppsDiv_32f_I(dev_direction_vec, device_scratch_padA,
                    num_of_particles));

        CHECK_CUDA_ERROR(nppsDiv_32f_I(dev_direction_vec, 
                    &device_scratch_padA[1 * num_of_particles], num_of_particles));

        CHECK_CUDA_ERROR(nppsDiv_32f_I(dev_direction_vec, 
                    &device_scratch_padA[2 * num_of_particles], num_of_particles));

        custom_kernel_process_ray_collision(num_of_particles, dev_direction_vec, 
                device_scratch_padA, device_velocity, device_position,
                array_size, &device_ray_direction[i*3]);

    }

    CHECK_CUDA_ERROR(cudaFree(dev_dot_product_result));
    CHECK_CUDA_ERROR(cudaFree(dev_direction_vec));

}


void universe::process_collisions_with_other_particles(void)
{
    
}

void universe::process_particles(void)
{
    const float delta = DELTA_TIME;
    
    // // Update the positions of the particles:
    // // self.positions[counti] = self.velocities[counti] * DELTA_TIME + self.positions[counti]
    CHECK_CUDA_ERROR(cublasSaxpy(cublas_handle, array_size*3, &delta,
                device_velocity, 1, device_position, 1));

    // Update the velocities of the particles:
    // self.velocities[counti] = self.accelerations[counti] * DELTA_TIME + self.velocities[counti]
    CHECK_CUDA_ERROR(cublasSaxpy(cublas_handle, array_size*3, &delta,
                device_acceleration, 1, device_velocity, 1));

    process_collisions_with_rays();

    process_collisions_with_other_particles();
}

void universe::get_positions(float ** in_out_buffer, int * size, int * stride)
{
    CHECK_CUDA_ERROR(cudaMemcpy(host_position, device_position, 
                long(array_size * 3 * sizeof(float)), cudaMemcpyDeviceToHost));

    *in_out_buffer = host_position;
    *size = num_of_particles;
    *stride = array_size;
}

universe::universe(int size)
{
curandCreateGenerator(&gen,
            CURAND_RNG_PSEUDO_DEFAULT);

    CHECK_CUDA_ERROR(cublasCreate(&cublas_handle));

    num_of_particles = size;
    array_size = size; 
    
    CHECK_CUDA_ERROR(cudaMallocManaged(&(device_position), 
                long(array_size*3*sizeof(float))));

    CHECK_CUDA_ERROR(cudaMallocManaged(&(device_velocity), 
                long(array_size*3*sizeof(float))));

    CHECK_CUDA_ERROR(cudaMallocManaged(&(device_acceleration),
                long(array_size*3*sizeof(float))));

    CHECK_CUDA_ERROR(cudaMallocManaged(&(device_scratch_padA),
                long(array_size*3*sizeof(float))));

    host_position = (float *)malloc(long(array_size*3*sizeof(float)));
    host_velocity = (float *)malloc(long(array_size*3*sizeof(float)));
    host_acceleration = (float *)malloc(long(array_size*3*sizeof(float)));

    CHECK_CUDA_ERROR(curandGenerateUniform(gen, device_position, array_size*2));
    CHECK_CUDA_ERROR(nppsAddC_32f_I(20.0, device_position, array_size*3));
    CHECK_CUDA_ERROR(curandGenerateUniform(gen, device_velocity, array_size*2));
    // CHECK_CUDA_ERROR(curandGenerateUniform(gen, device_acceleration, num_of_particles*3));

    //Add rays that can collied with particles to the universe
    CHECK_CUDA_ERROR(cudaMallocManaged(&(device_ray_position),
                long(4*3*sizeof(float))));

    CHECK_CUDA_ERROR(cudaMallocManaged(&(device_ray_direction),
                long(4*3*sizeof(float))));
    
    float temp_positions[] = {0.0,0.0,0.0,0.0,0.0,0.0,800.0,800.0,800.0,800.0,800.0,800.0};
    float temp_directions[] = {1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0, 0.0};

    CHECK_CUDA_ERROR(cudaMemcpy(device_ray_position, temp_positions,
                long(sizeof(temp_positions)), cudaMemcpyHostToDevice));

    CHECK_CUDA_ERROR(cudaMemcpy(device_ray_direction, temp_directions,
                long(sizeof(temp_directions)), cudaMemcpyHostToDevice));

    
}

universe::~universe()
{


    CHECK_CUDA_ERROR(cudaFree(device_position));
    CHECK_CUDA_ERROR(cudaFree(device_velocity));
    CHECK_CUDA_ERROR(cudaFree(device_acceleration));
    CHECK_CUDA_ERROR(cudaFree(device_scratch_padA));

    free(host_position);
    free(host_velocity);
    free(host_acceleration);
    CHECK_CUDA_ERROR(cublasDestroy(cublas_handle));



}

void universe::process(void)
{

    //First we need to process gravity



    //Then we want to process collisions
    process_particles();

}
